/* iRiver iFP 7XX Emulator
   Copyright (C) 2005 Tomasz Malesinski <tmal@mimuw.edu.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <stdlib.h>

#include "emuriver.h"
#include "cpu.h"
#include "log.h"

#define MODE_USER 0x10
#define MODE_FIQ  0x11
#define MODE_IRQ  0x12
#define MODE_SVC  0x13
#define MODE_ABT  0x17
#define MODE_UND  0x1b
#define MODE_SYS  0x1f

typedef void (*instruction)(struct cpu *, uint32);

static int breaks_find_ge(struct breaks *breaks, uint32 v) {
  int a, b;
  a = 0;
  b = breaks->n;
  while (a < b) {
    int m = (a + b) / 2;
    if (breaks->tab[m] >= v)
      b = m;
    else
      a = m + 1;
  }
  return a;
}

static void breaks_delete(struct breaks *breaks, uint32 v) {
  int i;
  i = breaks_find_ge(breaks, v);
  if (i < breaks->n && breaks->tab[i] == v) {
    int k;
    k = breaks->n - (i + 1);
    if (k)
      memmove(breaks->tab + i, breaks->tab + i + 1, k * sizeof(uint32));
    breaks->n--;
  }
}

static int breaks_insert(struct breaks *breaks, uint32 v) {
  int i, k;
  i = breaks_find_ge(breaks, v);
  if (i >= breaks->n || breaks->tab[i] != v) {
    if (breaks->n == breaks->max_n)
      return -1;

    k = breaks->n - i;
    if (k)
      memmove(breaks->tab + i + 1, breaks->tab + i, k * sizeof(uint32));

    breaks->tab[i] = v;
    breaks->n++;
  }
  return 0;
}

static int breaks_overlaps(struct breaks *breaks, uint32 a, uint32 b) {
  int i;
  i = breaks_find_ge(breaks, a);
  return (i < breaks->n && breaks->tab[i] <= b);
}

static int breaks_init(struct breaks *breaks, int max_n) {
  breaks->tab = malloc(sizeof(uint32) * max_n);
  if (!breaks->tab)
    return -1;

  breaks->n = 0;
  breaks->max_n = max_n;
  return 0;
}

static void breaks_cleanup(struct breaks *breaks) {
  free(breaks->tab);
}

static struct breaks *breaks_create(int max_n) {
  struct breaks *breaks;
  breaks = malloc(sizeof(struct breaks));
  if (!breaks)
    return NULL;

  if (breaks_init(breaks, max_n) < 0) {
    free(breaks);
    return NULL;
  }

  return breaks;
}

static struct breaks *cpu_select_breaks(struct cpu *cpu, int type) {
  switch (type) {
  case CPU_BRK_BREAK:
    return &cpu->breakpoints;
  case CPU_BRK_RWATCH:
    return &cpu->rwatchpoints;
  case CPU_BRK_WWATCH:
    return &cpu->wwatchpoints;
  case CPU_BRK_AWATCH:
    return &cpu->awatchpoints;
  default:
    return NULL;
  }
}

int cpu_insert_break(struct cpu *cpu, int type, uint32 val) {
  struct breaks *breaks = cpu_select_breaks(cpu, type);
  if (breaks)
    return breaks_insert(breaks, val);
  else
    return -1;
}

void cpu_remove_break(struct cpu *cpu, int type, uint32 val) {
  struct breaks *breaks = cpu_select_breaks(cpu, type);
  if (breaks)
    breaks_delete(breaks, val);
}

static void check_memory_read(struct cpu *cpu, uint32 addr1, uint32 addr2) {
  if (breaks_overlaps(&cpu->rwatchpoints, addr1, addr2)) {
    cpu->stop = CPU_STOP_RWATCH;
    cpu->break_addr = addr1;
  }

  if (breaks_overlaps(&cpu->awatchpoints, addr1, addr2)) {
    cpu->stop = CPU_STOP_AWATCH;
    cpu->break_addr = addr1;
  }
}

static void check_memory_write(struct cpu *cpu, uint32 addr1, uint32 addr2) {
  if (breaks_overlaps(&cpu->wwatchpoints, addr1, addr2)) {
    cpu->stop = CPU_STOP_WWATCH;
    cpu->break_addr = addr1;
  }

  if (breaks_overlaps(&cpu->awatchpoints, addr1, addr2)) {
    cpu->stop = CPU_STOP_AWATCH;
    cpu->break_addr = addr1;
  }
}

static uint8 cpu_memory_read(struct cpu *cpu, uint32 addr) {
  check_memory_read(cpu, addr, addr);
  return memory_read(cpu->mem, addr);
}

static uint32 cpu_memory_read_half_word(struct cpu *cpu, uint32 addr) {
  check_memory_read(cpu, addr, addr + 1);
  return memory_read_half_word(cpu->mem, addr);
}

static uint32 cpu_memory_read_word(struct cpu *cpu, uint32 addr) {
  check_memory_read(cpu, addr, addr + 3);
  return memory_read_word(cpu->mem, addr);
}

static void cpu_memory_write(struct cpu *cpu, uint32 addr, uint8 val) {
  check_memory_write(cpu, addr, addr);
  memory_write(cpu->mem, addr, val);
}

static void cpu_memory_write_half_word(struct cpu *cpu,
				       uint32 addr, uint32 val) {
  check_memory_write(cpu, addr, addr + 1);
  memory_write_half_word(cpu->mem, addr, val);
}

static void cpu_memory_write_word(struct cpu *cpu,
				  uint32 addr, uint32 val) {
  check_memory_write(cpu, addr, addr + 3);
  memory_write_word(cpu->mem, addr, val);
}

static void i_unknown(struct cpu *cpu, uint32 opcode) {
  cpu->stop = CPU_STOP_ILLEGAL;
  cpu->reg[15] -= 4;
}

static inline int is_extended(uint32 opcode) {
  return (opcode & 0x02000090) == 0x90;
}

static void set_n(struct cpu *cpu, int val) {
  cpu->cpsr = (cpu->cpsr & ~(1 << 31)) | (val << 31);
}

static void set_z(struct cpu *cpu, int val) {
  cpu->cpsr = (cpu->cpsr & ~(1 << 30)) | (val << 30);
}

static void set_nz(struct cpu *cpu, uint32 val) {
  cpu->cpsr = (cpu->cpsr & ~((1 << 31) | (1 << 30)))
    | (val & (1 << 31))
    | ((val == 0) << 30);
}

static void set_c(struct cpu *cpu, int val) {
  cpu->cpsr = (cpu->cpsr & ~(1 << 29)) | (val << 29);
}

static int get_c(struct cpu *cpu) {
  return bits(cpu->cpsr, 29, 1);
}

static void set_v_after_add(struct cpu *cpu, uint32 a, uint32 b, uint32 res) {
  int v = ((~a ^ b) & (a ^ res) & (1 << 31)) >> 3;
  cpu->cpsr = (cpu->cpsr & ~(1 << 28)) | v;
}

static void set_v_after_sub(struct cpu * cpu, uint32 a, uint32 b, uint32 res) {
  int v = ((a ^ b) & (a ^ res) & (1 << 31)) >> 3;
  cpu->cpsr = (cpu->cpsr & ~(1 << 28)) | v;
}

static uint32 shifter(struct cpu *cpu, uint32 opcode, int set_carry) {
  int b = bits(opcode, 20, 8);
  if ((b & 0xe0) == 0x20) {
    uint32 operand = ror(bits(opcode, 0, 8), bits(opcode, 8, 4) * 2);
    if (set_carry && bits(opcode, 8, 4) != 0)
      set_c(cpu, bits(operand, 31, 1));
    return operand;
  } else if ((b & 0xe0) == 0) {

    if (bits(opcode, 4, 3) == 0) {
      int shift_imm = bits(opcode, 7, 5);
      int m = bits(opcode, 0, 4);
      uint32 rm = cpu->reg[m] + (m == 15 ? 4 : 0);
      if (shift_imm == 0)
	return rm;
      else {
	if (set_carry)
	  set_c(cpu, bits(rm, 32 - shift_imm, 1));
	return rm << shift_imm;
      }

    } else if (bits(opcode, 4, 4) == 1) {
      int rs = cpu->reg[bits(opcode, 8, 4)] & 0xff;
      uint32 rm = cpu->reg[bits(opcode, 0, 4)];
      if (rs == 0)
	return rm;
      else if (rs < 32) {
	if (set_carry)
	  set_c(cpu, bits(rm, 32 - rs, 1));
	return rm << rs;
      } else if (rs == 32) {
	if (set_carry)
	  set_c(cpu, bits(rm, 0, 1));
	return 0;
      } else {
	if (set_carry)
	  set_c(cpu, 0);
	return 0;
      }

    } else if (bits(opcode, 4, 3) == 2) {
      int shift_imm = bits(opcode, 7, 5);
      int m = bits(opcode, 0, 4);
      uint32 rm = cpu->reg[m] + (m == 15 ? 4 : 0);
      if (shift_imm == 0) {
	if (set_carry)
	  set_c(cpu, bits(rm, 31, 1));
	return 0;
      } else {
	if (set_carry)
	  set_c(cpu, bits(rm, shift_imm - 1, 1));
	return rm >> shift_imm;
      }

    } else if (bits(opcode, 4, 4) == 3) {
      int rs = cpu->reg[bits(opcode, 8, 4)] & 0xff;
      uint32 rm = cpu->reg[bits(opcode, 0, 4)];
      if (rs == 0)
	return rm;
      else if (rs < 32) {
	if (set_carry)
	  set_c(cpu, bits(rm, rs - 1, 1));
	return rm >> rs;
      } else if (rs == 32) {
	if (set_carry)
	  set_c(cpu, bits(rm, 31, 1));
	return 0;
      } else {
	if (set_carry)
	  set_c(cpu, 0);
	return 0;
      }

    } else if (bits(opcode, 4, 3) == 4) {
      int shift_imm = bits(opcode, 7, 5);
      int m = bits(opcode, 0, 4);
      uint32 rm = cpu->reg[m] + (m == 15 ? 4 : 0);
      if (shift_imm == 0) {
	if (set_carry)
	  set_c(cpu, bits(rm, 31, 1));
	return bits(rm, 31, 1) ? ~0 : 0;
      } else {
	if (set_carry)
	  set_c(cpu, bits(rm, shift_imm - 1, 1));
	return asr(rm, shift_imm);
      }

    } else if (bits(opcode, 4, 4) == 5) {
      int rs = cpu->reg[bits(opcode, 8, 4)] & 0xff;
      uint32 rm = cpu->reg[bits(opcode, 0, 4)];
      if (rs == 0)
	return rm;
      else if (rs < 32) {
	if (set_carry)
	  set_c(cpu, bits(rm, rs - 1, 1));
	return asr(rm, rs);
      } else if (rs >= 32) {
	if (set_carry)
	  set_c(cpu, bits(rm, 31, 1));
	return bits(rm, 31, 1) ? ~0 : 0;
      }

    } else if (bits(opcode, 4, 3) == 6) {
      int shift_imm = bits(opcode, 7, 5);
      int m = bits(opcode, 0, 4);
      uint32 rm = cpu->reg[m] + (m == 15 ? 4 : 0);
      if (shift_imm == 0) {
	uint32 res = (rm >> 1) | (get_c(cpu) << 31);
	if (set_carry)
	  set_c(cpu, bits(rm, 0, 1));
	return res;

      } else {
	if (set_carry)
	  set_c(cpu, bits(rm, shift_imm - 1, 1));
	return ror(rm, shift_imm);
      }

    } else if (bits(opcode, 4, 4) == 7) {
      int rs = cpu->reg[bits(opcode, 8, 4)] & 0xff;
      uint32 rm = cpu->reg[bits(opcode, 0, 4)];
      if (rs == 0)
	return rm;
      else if ((rs & 0xf) == 0) {
	if (set_carry)
	  set_c(cpu, bits(rm, 31, 1));
	return rm;
      } else {
	int rs4 = rs & 0xf;
	if (set_carry)
	  set_c(cpu, bits(rm, rs4 - 1, 1));
	return ror(rm, rs4);
      }

    } else {
      fprintf(stderr, "Unknown shift %x at %x\n", opcode,
	      cpu->reg[15] - 4);
    }
  } else {
    fprintf(stderr, "Unknown addressing mode %x at %x\n", opcode,
	    cpu->reg[15] - 4);
    exit(1);
  }
}

static uint32 scaled_register_offset(struct cpu *cpu, uint32 opcode) {
    uint32 rm = cpu->reg[bits(opcode, 0, 4)];
    int shift_imm = bits(opcode, 7, 5);

    switch (bits(opcode, 5, 2)) {
    case 0:
      return rm << shift_imm;
    case 1:
      if (shift_imm == 0)
	return 0;
      else
	return rm >> shift_imm;
      break;
    case 2:
      if (shift_imm == 0)
	if (rm & (1 << 31))
	  return ~0;
	else
	  return 0;
      else
	return asr(rm, shift_imm);
      break;
    case 3:
      if (shift_imm == 0)
	return get_c(cpu) << 31 | (rm >> 1);
      else
	return ror(rm, shift_imm);
      break;
    }
}

static uint32 adrmode_2(struct cpu *cpu, uint32 opcode) {
  int b = bits(opcode, 20, 8);
  int n = bits(opcode, 16, 4);
  int rn;

  rn = cpu->reg[n];
  if (n == 15)
    rn += 4;

  if ((b & 0xfa) == 0x58) {
    /* immediate offset + */
    return rn + bits(opcode, 0, 12);

  } else if ((b & 0xfa) == 0x50) {
    /* immediate offset - */
    return rn - bits(opcode, 0, 12);

  } else if ((b & 0xf2) == 0x70) {
    /* scaled register offset */
    uint32 rm = cpu->reg[bits(opcode, 0, 4)];
    int shift_imm = bits(opcode, 7, 5);
    uint32 index = scaled_register_offset(cpu, opcode);

    if (opcode & (1 << 23))
      return rn + index;
    else
      return rn - index;

  } else if ((b & 0xf2) == 0x72) {
    /* scaled register pre-indexed */
    uint32 rm = cpu->reg[bits(opcode, 0, 4)];
    int shift_imm = bits(opcode, 7, 5);
    uint32 index = scaled_register_offset(cpu, opcode);

    if (opcode & (1 << 23))
      return cpu->reg[n] = rn + index;
    else
      return cpu->reg[n] = rn - index;

  } else if ((b & 0xf2) == 0x60) {
    /* scaled register post-indexed */
    uint32 rm = cpu->reg[bits(opcode, 0, 4)];
    int shift_imm = bits(opcode, 7, 5);
    uint32 index = scaled_register_offset(cpu, opcode);

    if (opcode & (1 << 23))
      cpu->reg[n] = rn + index;
    else
      cpu->reg[n] = rn - index;

    return rn;

  } else if ((b & 0xfa) == 0x48) {
    /* immediate post-indexed + */
    cpu->reg[n] = rn + bits(opcode, 0, 12);
    return rn;

  } else if ((b & 0xfa) == 0x40) {
    /* immediate post-indexed - */
    cpu->reg[n] = rn - bits(opcode, 0, 12);
    return rn;

  } else if ((b & 0xfa) == 0x5a) {
    /* immediate pre-indexed + */
    return cpu->reg[n] = rn + bits(opcode, 0, 12);

  } else if ((b & 0xfa) == 0x52) {
    /* immediate pre-indexed - */
    return cpu->reg[n] = rn - bits(opcode, 0, 12);

  } else {
    fprintf(stderr, "Unknown addressing mode %x at %x\n", opcode,
	    cpu->reg[15] - 4);
    exit(1);
  }
}

static uint32 adrmode_3(struct cpu *cpu, uint32 opcode) {
  int b = bits(opcode, 20, 8);
  int n = bits(opcode, 16, 4);
  int rn;

  rn = cpu->reg[n];
  if (n == 15)
    rn += 4;

  if ((b & 0xfe) == 0x1c) {
    /* immediate offset + */
    return rn + bits(opcode, 0, 4) + (bits(opcode, 8, 4) << 4);
  } else if ((b & 0xfe) == 0x14) {
    /* immediate offset - */
    return rn - bits(opcode, 0, 4) - (bits(opcode, 8, 4) << 4);
  } else if ((b & 0xfe) == 0x18) {
    /* register offset + */
    return rn + cpu->reg[bits(opcode, 0, 4)];
  } else if ((b & 0xfe) == 0x10) {
    /* register offset - */
    return rn - cpu->reg[bits(opcode, 0, 4)];

  } else if ((b & 0xfe) == 0x0c) {
    /* immediate post-indexed + */
    cpu->reg[n] += bits(opcode, 0, 4) + (bits(opcode, 8, 4) << 4);
    return rn;
  } else if ((b & 0xfe) == 0x04) {
    /* immediate post-indexed - */
    cpu->reg[n] -= bits(opcode, 0, 4) + (bits(opcode, 8, 4) << 4);
    return rn;
  } else if ((b & 0xfe) == 0x08) {
    /* register post-indexed + */
    cpu->reg[n] += cpu->reg[bits(opcode, 0, 4)];
    return rn;
  } else if ((b & 0xfe) == 0x00) {
    /* register post-indexed - */
    cpu->reg[n] -= cpu->reg[bits(opcode, 0, 4)];
    return rn;

  } else if ((b & 0xfe) == 0x1e) {
    /* immediate pre-indexed + */
    cpu->reg[n] += bits(opcode, 0, 4) + (bits(opcode, 8, 4) << 4);
    return rn;
  } else if ((b & 0xfe) == 0x16) {
    /* immediate pre-indexed - */
    cpu->reg[n] -= bits(opcode, 0, 4) + (bits(opcode, 8, 4) << 4);
    return rn;
  } else if ((b & 0xfe) == 0x1a) {
    /* register pre-indexed + */
    cpu->reg[n] += cpu->reg[bits(opcode, 0, 4)];
    return rn;
  } else if ((b & 0xfe) == 0x12) {
    /* register pre-indexed - */
    cpu->reg[n] -= cpu->reg[bits(opcode, 0, 4)];
    return rn;

  } else {
    fprintf(stderr, "Unknown addressing mode %x at %x\n", opcode,
	    cpu->reg[15] - 4);
    exit(1);
  }
}


static void i_ldr(struct cpu *cpu, uint32 opcode) {
  uint32 adr = adrmode_2(cpu, opcode);
  uint32 val = ror(cpu_memory_read_word(cpu, adr), adr & 3);
  int d = bits(opcode, 12, 4);
  if (d == 15)
    cpu->reg[15] = val & ~3;
  else
    cpu->reg[d] = val;
}

static void i_ldrb(struct cpu *cpu, uint32 opcode) {
  uint32 adr;
  uint8 val;

  if ((opcode & 0x0e000010) == 0x06000010) {
    i_unknown(cpu, opcode);
    return;
  }

  adr  = adrmode_2(cpu, opcode);
  val = cpu_memory_read(cpu, adr);
  int d = bits(opcode, 12, 4);
  cpu->reg[d] = val;
}

static inline void uint32_swap(uint32 *a, uint32 *b) {
  int t = *a;
  *a = *b;
  *b = t;
}

static void array_swap(uint32 *a, uint32 *b, int n) {
  int i;
  for (i = 0; i < n; i++)
    uint32_swap(a + i, b + i);
}

static void swap_regs(struct cpu *cpu, int mode) {
  int i;

  switch (mode) {
  case MODE_SVC:
    array_swap(cpu->reg + 13, cpu->saved_svc, 2);
    uint32_swap(&cpu->spsr, cpu->saved_svc + 2);
    break;

  case MODE_ABT:
    array_swap(cpu->reg + 13, cpu->saved_abt, 2);
    uint32_swap(&cpu->spsr, cpu->saved_abt + 2);
    break;

  case MODE_UND:
    array_swap(cpu->reg + 13, cpu->saved_und, 2);
    uint32_swap(&cpu->spsr, cpu->saved_und + 2);
    break;

  case MODE_IRQ:
    array_swap(cpu->reg + 13, cpu->saved_irq, 2);
    uint32_swap(&cpu->spsr, cpu->saved_irq + 2);
    break;

  case MODE_FIQ:
    array_swap(cpu->reg + 8, cpu->saved_fiq, 7);
    uint32_swap(&cpu->spsr, cpu->saved_fiq + 7);
    break;
  }
}

static inline set_cpsr(struct cpu *cpu, uint32 val) {
  swap_regs(cpu, cpu->cpsr & 0x1f);
  swap_regs(cpu, val & 0x1f);
  cpu->cpsr = val;
}

static void i_ldm1(struct cpu *cpu, uint32 opcode) {
  int n = bits(opcode, 16, 4);
  int id = bits(opcode, 23, 1);
  int ba = bits(opcode, 24, 1);
  uint32 adr = cpu->reg[n];
  int i;

  if (id) {
    if (ba) adr += 4;
    for (i = 0; i < 16; i++)
      if (opcode & (1 << i)) {
	cpu->reg[i] = cpu_memory_read_word(cpu, adr);
	adr += 4;
      }
    if (ba) adr -= 4;
  } else {
    if (ba) adr -= 4;
    for (i = 15; i >= 0; i--)
      if (opcode & (1 << i)) {
	cpu->reg[i] = cpu_memory_read_word(cpu, adr);
	adr -= 4;
      }
    if (ba) adr += 4;
  }

  if (opcode & (1 << 15))
    cpu->reg[15] &= ~3;

  if (opcode & (1 << 21))
    cpu->reg[n] = adr;
}

static void i_ldm3(struct cpu *cpu, uint32 opcode) {
  i_ldm1(cpu, opcode);
  set_cpsr(cpu, cpu->spsr);
}

static void i_ldm2_3(struct cpu *cpu, uint32 opcode) {
  int n = bits(opcode, 16, 4);
  int id = bits(opcode, 23, 1);
  int ba = bits(opcode, 24, 1);
  uint32 adr = cpu->reg[n];
  int i;

  if (bits(opcode, 15, 1)) {
    i_ldm3(cpu, opcode);
    return;
  }

  swap_regs(cpu, cpu->cpsr & 0x1f);
  if (id) {
    if (ba) adr += 4;
    for (i = 0; i < 16; i++)
      if (opcode & (1 << i)) {
	cpu->reg[i] = cpu_memory_read_word(cpu, adr);
	adr += 4;
      }
    if (ba) adr -= 4;
  } else {
    if (ba) adr -= 4;
    for (i = 15; i >= 0; i--)
      if (opcode & (1 << i)) {
	cpu->reg[i] = cpu_memory_read_word(cpu, adr);
	adr -= 4;
      }
    if (ba) adr += 4;
  }
  swap_regs(cpu, cpu->cpsr & 0x1f);
}

static void i_stm1(struct cpu *cpu, uint32 opcode) {
  int n = bits(opcode, 16, 4);
  int id = bits(opcode, 23, 1);
  int ba = bits(opcode, 24, 1);
  uint32 adr = cpu->reg[n];
  int i;

  if (id) {
    if (ba) adr += 4;
    for (i = 0; i < 15; i++)
      if (opcode & (1 << i)) {
	cpu_memory_write_word(cpu, adr, cpu->reg[i]);
	adr += 4;
      }
    if (ba) adr -= 4;
  } else {
    if (ba) adr -= 4;
    for (i = 15; i >= 0; i--)
      if (opcode & (1 << i)) {
	cpu_memory_write_word(cpu, adr, cpu->reg[i]);
	adr -= 4;
      }
    if (ba) adr += 4;
  }

  if (opcode & (1 << 21))
    cpu->reg[n] = adr;
}

static void i_stm2(struct cpu *cpu, uint32 opcode) {
  int n = bits(opcode, 16, 4);
  int id = bits(opcode, 23, 1);
  int ba = bits(opcode, 24, 1);
  uint32 adr = cpu->reg[n];
  int i;

  swap_regs(cpu, cpu->cpsr & 0x1f);
  if (id) {
    if (ba) adr += 4;
    for (i = 0; i < 15; i++)
      if (opcode & (1 << i)) {
	cpu_memory_write_word(cpu, adr, cpu->reg[i]);
	adr += 4;
      }
    if (ba) adr -= 4;
  } else {
    if (ba) adr -= 4;
    for (i = 15; i >= 0; i--)
      if (opcode & (1 << i)) {
	cpu_memory_write_word(cpu, adr, cpu->reg[i]);
	adr -= 4;
      }
    if (ba) adr += 4;
  }
  swap_regs(cpu, cpu->cpsr & 0x1f);
}

static inline int privileged_mode(struct cpu *cpu) {
  return (cpu->cpsr & 0x1f) != MODE_USER;
}

static inline int mode_has_spsr(struct cpu *cpu) {
  return (cpu->cpsr & 0x1f) != MODE_USER
    && (cpu->cpsr & 0x1f) != MODE_SYS;
}

void cpu_set_irq_p(struct cpu *cpu, int (*irq_p)(void *),
		   void *irq_data) {
  cpu->irq_p = irq_p;
  cpu->irq_data = irq_data;
}

void cpu_irq(struct cpu *cpu) {
  swap_regs(cpu, cpu->cpsr & 0x1f);
  swap_regs(cpu, 0x12);

  cpu->spsr = cpu->cpsr;
  cpu->reg[14] = cpu->reg[15] + 4;
  cpu->reg[15] = 0x18;
  cpu->cpsr = (cpu->cpsr & ~0xbf) | 0x92;
}

static void i_str(struct cpu *cpu, uint32 opcode) {
  /* TODO what if rd=15? */
  uint32 adr = adrmode_2(cpu, opcode);
  cpu_memory_write_word(cpu, adr, cpu->reg[bits(opcode, 12, 4)]);
}

static void i_strb(struct cpu *cpu, uint32 opcode) {
  uint32 adr = adrmode_2(cpu, opcode);
  cpu_memory_write(cpu, adr, cpu->reg[bits(opcode, 12, 4)] & 0xff);
}

static void i_mul(struct cpu *cpu, uint32 opcode) {
  uint32 res = cpu->reg[bits(opcode, 0, 4)] * cpu->reg[bits(opcode, 8, 4)];
  cpu->reg[bits(opcode, 16, 4)] = res;
  if (opcode & (1 << 20))
    set_nz(cpu, res);
}

static void i_mla(struct cpu *cpu, uint32 opcode) {
  uint32 res = cpu->reg[bits(opcode, 0, 4)] * cpu->reg[bits(opcode, 8, 4)]
    + cpu->reg[bits(opcode, 12, 4)];
  cpu->reg[bits(opcode, 16, 4)] = res;
  if (opcode & (1 << 20))
    set_nz(cpu, res);
}

static void i_umull(struct cpu *cpu, uint32 opcode) {
  uint64 res = (uint64)cpu->reg[bits(opcode, 0, 4)] * cpu->reg[bits(opcode, 8, 4)];
  cpu->reg[bits(opcode, 12, 4)] = res & 0xffffffff;
  cpu->reg[bits(opcode, 16, 4)] = res >> 32;
  if (opcode & (1 << 20)) {
    set_n(cpu, (res & (1ULL << 63)) != 0);
    set_z(cpu, res == 0);
  }
}

static void i_smull(struct cpu *cpu, uint32 opcode) {
  sint64 res = (sint64)((sint32)cpu->reg[bits(opcode, 0, 4)])
    * cpu->reg[bits(opcode, 8, 4)];
  cpu->reg[bits(opcode, 12, 4)] = res & 0xffffffff;
  cpu->reg[bits(opcode, 16, 4)] = res >> 32;
  if (opcode & (1 << 20)) {
    set_n(cpu, res < 0);
    set_z(cpu, res == 0);
  }
}

static void i_umlal(struct cpu *cpu, uint32 opcode) {
  int rdl = bits(opcode, 12, 4);
  int rdh = bits(opcode, 16, 4);
  uint64 res = (uint64)cpu->reg[bits(opcode, 0, 4)] * cpu->reg[bits(opcode, 8, 4)]
    + ((uint64)cpu->reg[rdh] << 32) + cpu->reg[rdl];
  cpu->reg[rdl] = res & 0xffffffff;
  cpu->reg[rdh] = res >> 32;
  if (opcode & (1 << 20)) {
    set_n(cpu, (res & (1ULL << 63)) != 0);
    set_z(cpu, res == 0);
  }
}

static void i_smlal(struct cpu *cpu, uint32 opcode) {
  int rdl = bits(opcode, 12, 4);
  int rdh = bits(opcode, 16, 4);
  sint64 res = (sint64)((sint32)cpu->reg[bits(opcode, 0, 4)])
    * cpu->reg[bits(opcode, 8, 4)]
    + ((sint64)cpu->reg[rdh] << 32) + cpu->reg[rdl];
  cpu->reg[rdl] = res & 0xffffffff;
  cpu->reg[rdh] = res >> 32;
  if (opcode & (1 << 20)) {
    set_n(cpu, res < 0);
    set_z(cpu, res == 0);
  }
}

static void i_strh(struct cpu *cpu, uint32 opcode) {
  uint32 data = cpu->reg[bits(opcode, 12, 4)];
  uint32 adr = adrmode_3(cpu, opcode);
  cpu_memory_write_half_word(cpu, adr, data & 0xffff);
}

static void i_ldrh(struct cpu *cpu, uint32 opcode) {
  uint32 adr = adrmode_3(cpu, opcode);
  uint32 data = cpu_memory_read_half_word(cpu, adr);
  cpu->reg[bits(opcode, 12, 4)] = data;
}

static void i_ldrsh(struct cpu *cpu, uint32 opcode) {
  uint32 adr = adrmode_3(cpu, opcode);
  uint32 data = cpu_memory_read_half_word(cpu, adr);
  cpu->reg[bits(opcode, 12, 4)] = sign_extend(data, 16);
}

static void i_ldrsb(struct cpu *cpu, uint32 opcode) {
  uint32 adr = adrmode_3(cpu, opcode);
  uint32 data = cpu_memory_read(cpu, adr);
  cpu->reg[bits(opcode, 12, 4)] = sign_extend(data, 8);
}

static void i_extended(struct cpu *cpu, uint32 opcode) {
  switch (opcode & 0x0ff000f0) {
  case 0x90:
  case 0x00100090:
    i_mul(cpu, opcode);
    break;
  case 0x00200090:
  case 0x00300090:
    i_mla(cpu, opcode);
    break;
  case 0x00800090:
  case 0x00900090:
    i_umull(cpu, opcode);
    break;
  case 0x00a00090:
  case 0x00b00090:
    i_umlal(cpu, opcode);
    break;
  case 0x00c00090:
  case 0x00d00090:
    i_smull(cpu, opcode);
    break;
  case 0x00e00090:
  case 0x00f00090:
    i_smlal(cpu, opcode);
    break;
  default:
    switch (opcode & 0x0e1000f0) {
    case 0x000000b0:
      i_strh(cpu, opcode);
      break;
    case 0x001000b0:
      i_ldrh(cpu, opcode);
      break;
    case 0x001000d0:
      i_ldrsb(cpu, opcode);
      break;
    case 0x001000f0:
      i_ldrsh(cpu, opcode);
      break;
    default:
      i_unknown(cpu, opcode);
      break;
    }
  }
}

static void i_adc(struct cpu *cpu, uint32 opcode) {
  int s, d, n;
  uint32 val, rn;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  s = bits(opcode, 20, 1);
  d = bits(opcode, 12, 4);
  val = shifter(cpu, opcode, 0);
  n = bits(opcode, 16, 4);
  rn = cpu->reg[n];
  if (n == 15) rn += 4;
  unsigned long long resl = (unsigned long long)rn + val + get_c(cpu);
  uint32 res = resl & 0xffffffff;
  int c = (resl >> 32) & 1;
  cpu->reg[d] = res;
  if (s) {
    if (d == 15)
      set_cpsr(cpu, cpu->spsr);
    else {
      set_nz(cpu, res);
      set_c(cpu, c);
      set_v_after_add(cpu, rn, val, res);
    }
  }
}

static void i_add(struct cpu *cpu, uint32 opcode) {
  int s, d, n;
  uint32 val, rn;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  s = bits(opcode, 20, 1);
  d = bits(opcode, 12, 4);
  val = shifter(cpu, opcode, 0);
  n = bits(opcode, 16, 4);
  rn = cpu->reg[n];
  if (n == 15) rn += 4;
  unsigned long long resl = (unsigned long long)rn + val;
  uint32 res = resl & 0xffffffff;
  int c = (resl >> 32) & 1;
  cpu->reg[d] = res;
  if (s) {
    if (d == 15)
      set_cpsr(cpu, cpu->spsr);
    else {
      set_nz(cpu, res);
      set_c(cpu, c);
      set_v_after_add(cpu, rn, val, res);
    }
  }
}

static void i_and(struct cpu *cpu, uint32 opcode) {
  int d, s, n;
  uint32 val, rn, res;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  d = bits(opcode, 12, 4);
  s = bits(opcode, 20, 1);
  val = shifter(cpu, opcode, s);
  n = bits(opcode, 16, 4);
  rn = cpu->reg[n];
  if (n == 15) rn += 4;
  res = rn & val;
  cpu->reg[d] = res;
  if (s)
    if (d == 15)
      set_cpsr(cpu, cpu->spsr);
    else
      set_nz(cpu, res);
}

static void i_bic(struct cpu *cpu, uint32 opcode) {

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  int d = bits(opcode, 12, 4);
  int s = bits(opcode, 20, 1);
  uint32 val = shifter(cpu, opcode, s);
  int n = bits(opcode, 16, 4);
  uint32 rn = cpu->reg[n];
  if (n == 15) rn += 4;
  uint32 res = rn & ~val;
  cpu->reg[d] = res;
  if (s)
    if (d == 15)
      set_cpsr(cpu, cpu->spsr);
    else
      set_nz(cpu, res);
}

static void i_cmp(struct cpu *cpu, uint32 opcode) {
  uint32 val, rn;
  int n;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  val = shifter(cpu, opcode, 0);
  n = bits(opcode, 16, 4);
  rn = cpu->reg[n];
  if (n == 15) rn += 4;
  uint32 res = rn - val;
  set_nz(cpu, res);
  set_c(cpu, rn >= val);
  set_v_after_sub(cpu, rn, val, rn - val);
}

static void i_cmn(struct cpu *cpu, uint32 opcode) {
  uint32 val, rn;
  int n;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  val = shifter(cpu, opcode, 0);
  n = bits(opcode, 16, 4);
  rn = cpu->reg[n];
  if (n == 15) rn += 4;

  unsigned long long resl = (unsigned long long)rn + val;
  uint32 res = resl & 0xffffffff;

  set_nz(cpu, res);
  set_c(cpu, (resl >> 32) & 1);
  set_v_after_add(cpu, rn, val, res);
}

static void i_eor(struct cpu *cpu, uint32 opcode) {
  int d, s, n;
  uint32 val, rn, res;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  d = bits(opcode, 12, 4);
  s = bits(opcode, 20, 1);
  val = shifter(cpu, opcode, s);
  n = bits(opcode, 16, 4);
  rn = cpu->reg[n];
  if (n == 15) rn += 4;
  res = rn ^ val;
  cpu->reg[d] = res;
  if (s)
    if (d == 15)
      set_cpsr(cpu, cpu->spsr);
    else
      set_nz(cpu, res);
}

static void i_mov(struct cpu *cpu, uint32 opcode) {
  int d, s;
  uint32 val;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  d = bits(opcode, 12, 4);
  s = bits(opcode, 20, 1);
  val = shifter(cpu, opcode, s);
  cpu->reg[d] = val;
  if (s)
    if (d == 15)
      set_cpsr(cpu, cpu->spsr);
    else
      set_nz(cpu, val);
}

static void i_mvn(struct cpu *cpu, uint32 opcode) {
  int d, s;
  uint32 val;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  d = bits(opcode, 12, 4);
  s = bits(opcode, 20, 1);
  val = ~shifter(cpu, opcode, s);
  cpu->reg[d] = val;
  if (s)
    if (d == 15)
      set_cpsr(cpu, cpu->spsr);
    else
      set_nz(cpu, val);
}

static void i_orr(struct cpu *cpu, uint32 opcode) {
  int d, s, n;
  uint32 val, rn;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  d = bits(opcode, 12, 4);
  s = bits(opcode, 20, 1);
  val = shifter(cpu, opcode, s);
  n = bits(opcode, 16, 4);
  rn = cpu->reg[n];
  if (n == 15) rn += 4;
  uint32 res = rn | val;
  cpu->reg[d] = res;
  if (s)
    if (d == 15)
      set_cpsr(cpu, cpu->spsr);
    else
      set_nz(cpu, res);
}

static void i_rsb(struct cpu *cpu, uint32 opcode) {
  int s, d, n;
  uint32 val, rn;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  s = bits(opcode, 20, 1);
  d = bits(opcode, 12, 4);
  val = shifter(cpu, opcode, 0);
  n = bits(opcode, 16, 4);
  rn = cpu->reg[n];
  if (n == 15) rn += 4;
  unsigned long long resl = (unsigned long long)val - rn;
  uint32 res = resl & 0xffffffff;
  int c = ~(resl >> 32) & 1;
  cpu->reg[d] = res;
  if (s) {
    if (d == 15)
      set_cpsr(cpu, cpu->spsr);
    else {
      set_nz(cpu, res);
      set_c(cpu, c);
      set_v_after_sub(cpu, val, rn, res);
    }
  }
}

static void i_rsc(struct cpu *cpu, uint32 opcode) {
  int s, d, n;
  uint32 val, rn;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  s = bits(opcode, 20, 1);
  d = bits(opcode, 12, 4);
  val = shifter(cpu, opcode, 0);
  n = bits(opcode, 16, 4);
  rn = cpu->reg[n];
  if (n == 15) rn += 4;
  unsigned long long resl = (unsigned long long)val - rn + get_c(cpu) - 1;
  uint32 res = resl & 0xffffffff;
  int c = ~(resl >> 32) & 1;
  cpu->reg[d] = res;
  if (s) {
    if (d == 15)
      set_cpsr(cpu, cpu->spsr);
    else {
      set_nz(cpu, res);
      set_c(cpu, c);
      set_v_after_sub(cpu, val, rn, res);
    }
  }
}

static void i_sbc(struct cpu *cpu, uint32 opcode) {
  int s, d, n;
  uint32 val, rn;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  s = bits(opcode, 20, 1);
  d = bits(opcode, 12, 4);
  val = shifter(cpu, opcode, 0);
  n = bits(opcode, 16, 4);
  rn = cpu->reg[n];
  if (n == 15) rn += 4;
  unsigned long long resl = (unsigned long long)rn - val + get_c(cpu) - 1;
  uint32 res = resl & 0xffffffff;
  int c = ~(resl >> 32) & 1;
  cpu->reg[d] = res;
  if (s) {
    if (d == 15)
      set_cpsr(cpu, cpu->spsr);
    else {
      set_nz(cpu, res);
      set_c(cpu, c);
      set_v_after_sub(cpu, rn, val, res);
    }
  }
}

static void i_sub(struct cpu *cpu, uint32 opcode) {
  int s, d, n;
  uint32 val, rn;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  s = bits(opcode, 20, 1);
  d = bits(opcode, 12, 4);
  val = shifter(cpu, opcode, 0);
  n = bits(opcode, 16, 4);
  rn = cpu->reg[n];
  if (n == 15) rn += 4;
  unsigned long long resl = (unsigned long long)rn - val;
  uint32 res = resl & 0xffffffff;
  int c = ~(resl >> 32) & 1;
  cpu->reg[d] = res;
  if (s) {
    if (d == 15)
      set_cpsr(cpu, cpu->spsr);
    else {
      set_nz(cpu, res);
      set_c(cpu, c);
      set_v_after_sub(cpu, rn, val, res);
    }
  }
}

static void i_teq(struct cpu *cpu, uint32 opcode) {
  int n;
  uint32 val, rn;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  val = shifter(cpu, opcode, 1);
  n = bits(opcode, 16, 4);
  rn = cpu->reg[n];
  if (n == 15) rn += 4;
  set_nz(cpu, val ^ rn);
}

static void i_tst(struct cpu *cpu, uint32 opcode) {
  int n;
  uint32 val, rn;

  if (is_extended(opcode)) {
    i_extended(cpu, opcode);
    return;
  }

  val = shifter(cpu, opcode, 1);
  n = bits(opcode, 16, 4);
  rn = cpu->reg[n];
  if (n == 15) rn += 4;
  set_nz(cpu, val & rn);
}

static void i_b(struct cpu *cpu, uint32 opcode) {
  int offs = sign_extend(bits(opcode, 0, 24), 24) << 2;
  cpu->reg[15] += offs + 4;
}

static void i_bl(struct cpu *cpu, uint32 opcode) {
  int offs = sign_extend(bits(opcode, 0, 24), 24) << 2;
  cpu->reg[14] = cpu->reg[15];
  cpu->reg[15] += offs + 4;
}

static void i_mrs(struct cpu *cpu, uint32 opcode) {
  int d = bits(opcode, 12, 4);
  if (opcode & (1 << 22))
    cpu->reg[d] = cpu->spsr;
  else
    cpu->reg[d] = cpu->cpsr;
}

static void i_msr(struct cpu *cpu, uint32 opcode) {
  uint32 operand;
  int i;

  if (opcode & (1 << 25))
    operand = ror(bits(opcode, 0, 8), bits(opcode, 8, 4) * 2);
  else
    operand = cpu->reg[bits(opcode, 0, 4)];

  if (!(opcode & (1 << 22))) {
    if (privileged_mode(cpu))
      for (i = 0; i < 3; i++)
	if (opcode & (1 << (16 + i))) {
	  if (i == 0) {
	    swap_regs(cpu, cpu->cpsr & 0x1f);
	    swap_regs(cpu, operand & 0x1f);
	  }
	  cpu->cpsr = (cpu->cpsr & ~(0xff << (i * 8))) | (operand & (0xff << (i * 8)));
	}

    if (opcode & (1 << 19))
      cpu->cpsr = (cpu->cpsr & ~(0xff << 24)) | (operand & (0xff << 24));

  } else if (mode_has_spsr(cpu))
    for (i = 0; i < 4; i++)
      if (opcode & (1 << (16 + i)))
	cpu->spsr = (cpu->cpsr & ~(0xff << (i * 8))) | (operand & (0xff << (i * 8)));
}

static instruction instruction_table[256] = {
  i_and,     i_and,     i_eor,     i_eor,
  i_sub,     i_sub,     i_rsb,     i_rsb,
  i_add,     i_add,     i_adc,     i_adc,
  i_sbc,     i_sbc,     i_rsc,     i_rsc,
  
  i_mrs,     i_tst,     i_msr,     i_teq,
  i_mrs,     i_cmp,     i_msr,     i_cmn,
  i_orr,     i_orr,     i_mov,     i_mov,
  i_bic,     i_bic,     i_mvn,     i_mvn,

  i_and,     i_and,     i_eor,     i_eor,
  i_sub,     i_sub,     i_rsb,     i_rsb,
  i_add,     i_add,     i_adc,     i_adc,
  i_sbc,     i_sbc,     i_rsc,     i_rsc,

  i_unknown, i_tst,     i_msr,     i_teq,
  i_unknown, i_cmp,     i_msr,     i_cmn,
  i_orr,     i_orr,     i_mov,     i_mov,
  i_bic,     i_bic,     i_mvn,     i_mvn,

  i_str,     i_ldr,     i_str,     i_ldr,
  i_strb,    i_ldrb,    i_strb,    i_ldrb,
  i_str,     i_ldr,     i_str,     i_ldr,
  i_strb,    i_ldrb,    i_strb,    i_ldrb,

  i_str,     i_ldr,     i_str,     i_ldr,
  i_strb,    i_ldrb,    i_strb,    i_ldrb,
  i_str,     i_ldr,     i_str,     i_ldr,
  i_strb,    i_ldrb,    i_strb,    i_ldrb,

  i_str,     i_ldr,     i_str,     i_ldr,
  i_strb,    i_ldrb,    i_strb,    i_ldrb,
  i_str,     i_ldr,     i_str,     i_ldr,
  i_strb,    i_ldrb,    i_strb,    i_ldrb,

  i_str,     i_ldr,     i_str,     i_ldr,
  i_strb,    i_ldrb,    i_strb,    i_ldrb,
  i_str,     i_ldr,     i_str,     i_ldr,
  i_strb,    i_ldrb,    i_strb,    i_ldrb,

  i_stm1,    i_ldm1,    i_stm1,    i_ldm1,
  i_stm2,    i_ldm2_3,  i_unknown, i_unknown,
  i_stm1,    i_ldm1,    i_stm1,    i_ldm1,
  i_stm2,    i_ldm2_3,  i_unknown, i_unknown,

  i_stm1,    i_ldm1,    i_stm1,    i_ldm1,
  i_stm2,    i_ldm2_3,  i_unknown, i_unknown,
  i_stm1,    i_ldm1,    i_stm1,    i_ldm1,
  i_stm2,    i_ldm2_3,  i_unknown, i_unknown,

  i_b,       i_b,       i_b,       i_b,
  i_b,       i_b,       i_b,       i_b,
  i_b,       i_b,       i_b,       i_b,
  i_b,       i_b,       i_b,       i_b,

  i_bl,      i_bl,      i_bl,      i_bl,
  i_bl,      i_bl,      i_bl,      i_bl,
  i_bl,      i_bl,      i_bl,      i_bl,
  i_bl,      i_bl,      i_bl,      i_bl,

  i_unknown, i_unknown, i_unknown, i_unknown,
  i_unknown, i_unknown, i_unknown, i_unknown,
  i_unknown, i_unknown, i_unknown, i_unknown,
  i_unknown, i_unknown, i_unknown, i_unknown,

  i_unknown, i_unknown, i_unknown, i_unknown,
  i_unknown, i_unknown, i_unknown, i_unknown,
  i_unknown, i_unknown, i_unknown, i_unknown,
  i_unknown, i_unknown, i_unknown, i_unknown,

  i_unknown, i_unknown, i_unknown, i_unknown,
  i_unknown, i_unknown, i_unknown, i_unknown,
  i_unknown, i_unknown, i_unknown, i_unknown,
  i_unknown, i_unknown, i_unknown, i_unknown,

  i_unknown, i_unknown, i_unknown, i_unknown,
  i_unknown, i_unknown, i_unknown, i_unknown,
  i_unknown, i_unknown, i_unknown, i_unknown,
  i_unknown, i_unknown, i_unknown, i_unknown
};


struct cpu *create_cpu(struct memory *memory) {
  struct cpu *cpu;

  cpu = malloc(sizeof(struct cpu));
  if (!cpu) return NULL;

  if (breaks_init(&cpu->breakpoints, 1024) < 0)
    goto nobr;
  if (breaks_init(&cpu->rwatchpoints, 1024) < 0)
    goto norw;
  if (breaks_init(&cpu->wwatchpoints, 1024) < 0)
    goto noww;
  if (breaks_init(&cpu->awatchpoints, 1024) < 0)
    goto noaw;

  cpu->mem = memory;
  cpu->cpsr = 0xd3;
  cpu->irq_p = NULL;
  return cpu;

 noaw:
  breaks_cleanup(&cpu->wwatchpoints);
 noww:
  breaks_cleanup(&cpu->rwatchpoints);
 norw:
  breaks_cleanup(&cpu->breakpoints);
 nobr:
  free(cpu);
 err:
  error("error while creating cpu data structures (probably no memory)");
  return NULL;
}

static inline get_condition(struct cpu *cpu, int cond) {
  int b;
  switch (cond) {
  case 0: /* eq */
    return bits(cpu->cpsr, 30, 1);
  case 1: /* ne */
    return !bits(cpu->cpsr, 30, 1);
  case 2: /* cs */
    return get_c(cpu);
  case 3: /* cc */
    return !get_c(cpu);
  case 4: /* mi */
    return bits(cpu->cpsr, 31, 1);
  case 5: /* pl */
    return !bits(cpu->cpsr, 31, 1);
  case 6: /* vs */
    return bits(cpu->cpsr, 28, 1);
  case 7: /* vc */
    return !bits(cpu->cpsr, 28, 1);
  case 8: /* hi */
    return (bits(cpu->cpsr, 28, 4) & 6) == 2;
  case 9: /* ls */
    return (bits(cpu->cpsr, 28, 4) & 6) != 2;
  case 10: /* ge */
    b = bits(cpu->cpsr, 28, 4) & 9;
    return b == 9 || b == 0;
  case 11: /* lt */
    b = bits(cpu->cpsr, 28, 4) & 9;
    return b == 1 || b == 8;
  case 12: /* gt */
    b = bits(cpu->cpsr, 28, 4) & 13;
    return b == 9 || b == 0;
  case 13: /* le */
    b = bits(cpu->cpsr, 28, 4) & 13;
    return !(b == 9 || b == 0);
  case 14:
    return 1;
  case 15:
    return 0;
  }
}

void cpu_print_registers(struct cpu *cpu, FILE *fp) {
  int i;
  for (i = 0; i < 16; i++)
    fprintf(fp, "r%-2d=%8x%c", i, cpu->reg[i], ((i % 4) == 3) ? '\n' : ' ');
}

void cpu_execute_instruction(struct cpu *cpu) {
  uint32 instr;
  static int dbg = 0;
  uint32 old_pc;

  if (dbg)
    cpu_print_registers(cpu, stderr);

  if (breaks_overlaps(&cpu->breakpoints,
		      cpu->reg[15], cpu->reg[15] + 3)) {
    cpu->stop = CPU_STOP_BREAK;
    cpu->break_addr = cpu->reg[15];
    return;
  }

  instr = cpu_memory_read_word(cpu, cpu->reg[15]);
  cpu->reg[15] += 4;
  if (get_condition(cpu, instr >> 28))
    (*(instruction_table[bits(instr, 20, 8)]))(cpu, instr);
}

int cpu_go(struct cpu *cpu, int n_instr) {
  cpu->stop = 0;
  while (n_instr-- > 0) {
    if (!(cpu->cpsr & 0x80) && cpu->irq_p) {
      if ((*(cpu->irq_p))(cpu->irq_data))
	cpu_irq(cpu);
    }

    trace(cpu);
    cpu_execute_instruction(cpu);
    if (cpu->stop)
      return cpu->stop;
  }
  return CPU_STOP_NORMAL;
}
