/* iRiver iFP 7XX Emulator
   Copyright (C) 2005 Tomasz Malesinski <tmal@mimuw.edu.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <getopt.h>
#include <errno.h>

#include "emuriver.h"
#include "memory.h"
#include "cpu.h"
#include "hardware.h"
#include "trace.h"
#include "trainer.h"
#include "gdb.h"
#include "ui.h"

struct memory *memory;
struct cpu *cpu;
struct hardware *hardware;

struct mem_block *ram0, *ram4, *ram24;

char *program_name;

int option_gdb = 0;
int option_disable_ui = 0;
char *log_file = NULL;
char *flash_file = "flash.img";
char *prog_flash_file = "prog_flash.bin";
char *firmware1_file = "ifp1.bin";
char *firmware2_file = "ifp2.bin";
char *rockbox_file = NULL;

struct option options[] = {
  {"log", required_argument, NULL, 'l'},
  {"gdb", no_argument, NULL, 'g'},
  {"disable-ui", no_argument, NULL, 'n'},
  {"flash-image", required_argument, NULL, 'f'},
  {"prog-flash-image", required_argument, NULL, 'p'},
  {"firmware1", required_argument, NULL, '1'},
  {"firmware2", required_argument, NULL, '2'},
  {"rockbox", required_argument, NULL, 'r'},
};

void error(const char *format, ...) {
  va_list ap;
  fprintf(stderr, "%s: ", program_name);
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  fputc('\n', stderr);
}

void error_sys(const char *format, ...) {
  va_list ap;
  fprintf(stderr, "%s: ", program_name);
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  fprintf(stderr, "%s\n", strerror(errno));
}

void *xmalloc(size_t size) {
  void *p = malloc(size);
  if (!p)
    error("out of memory");
  return p;
}

int set_iriver_memory(struct memory *memory) {
  FILE *fp;
  ram0 = create_memory_area_ram(memory, 0, 0x10000);
  ram4 = create_memory_area_ram(memory, 0x400000, 0x10000);
  ram24 = create_memory_area_ram(memory, 0x24000000, 0x100000);
  if (!ram0 || !ram4 || !ram24)
    return -1;
  if (!create_memory_area_rom(memory, 0x800000, 0x80000, firmware1_file))
    return -1;
  if (!create_memory_area_rom(memory, 0xa00000, 0x100000, firmware2_file))
    return -1;

  if (rockbox_file) {
    fp = fopen(rockbox_file, "rb");
    if (!fp) {
      error_sys("%s: ", rockbox_file);
      return -1;
    }
    fread(ram24->mem, 1, 0x80000, fp);
    fclose(fp);
  }
  return 0;
}

int set_mpio_memory(struct memory *memory) {
  ram0 = create_memory_area_ram(memory, 0, 0x10000);
  ram4 = create_memory_area_ram(memory, 0x20010000, 0x20000);
  if (!ram0 || !ram4)
    return -1;
  return 0;
}

int save_ram(struct mem_block *mem, char *file) {
  FILE *fp;
  fp = fopen(file, "wb");
  if (!fp) return -1;
  fwrite(mem->mem, 1, 0x10000, fp);
  fclose(fp);

  return 0;
}

int load_binary(char *file, void *ptr, size_t len) {
  FILE *fp;
  fp = fopen(file, "rb");
  if (!fp) return -1;
  fread(ptr, 1, len, fp);
  fclose(fp);
}

int init_iriver(struct cpu *cpu, struct memory *memory, int upgrade) {
  if (set_iriver_memory(memory) < 0)
    return -1;

  if (upgrade) {
    load_binary("fw_upgrade.bin", ram0->mem + 0x80, 0x10000);
    cpu->reg[15] = 0x000080;
  } else if (rockbox_file)
    cpu->reg[15] = 0x24000000;
  else
    cpu->reg[15] = 0x800000;
  return 0;
}

void init_mpio(struct cpu *cpu, struct memory *memory, int upgrade) {
  set_mpio_memory(memory);
  if (upgrade) {
    load_binary("mpio_upgrade.bin", ram0->mem + 0x80, 0x10000);
    cpu->reg[15] = 0x000080;
  } else
    cpu->reg[15] = 0x800000;
}

int go(int n_cycles, int (*f)(void), int freq) {
  int i, stop, r;

  i = 0;
  while (n_cycles == 0 || i < n_cycles) {
    stop = cpu_go(cpu, 1);
    tick++;
    
    handle_events(tick);

    /*
    if ((tick % 1000000) == 5000) {
      cpu_irq(cpu);
      set_irq_no(0x1c);
    }
    */

    if (stop != CPU_STOP_NORMAL)
      return stop;

    if (f && (tick % freq) == 0) {
      r = (*f)();
      if (r)
	return r;
    }

    if (!option_disable_ui && (tick % 100000) == 0) {
      ui_refresh(hardware->lcd);
      if (ui_get_key(0)) return CPU_STOP_QUIT;
    }

    i++;
  }

  return CPU_STOP_NORMAL;
}

int parse_options(int argc, char *argv[]) {
  int c, index;
  FILE *log_fp;

  while (1) {
    c = getopt_long(argc, argv, "l:gnf:p:1:2:", options, &index);
    if (c == -1)
      break;

    switch (c) {
    case '?':
      return -1;
    case 'l':
      log_file = optarg;
      break;
    case 'g':
      option_gdb = 1;
      break;
    case 'n':
      option_disable_ui = 1;
      break;
    case 'f':
      flash_file = optarg;
      break;
    case 'p':
      prog_flash_file = optarg;
      break;
    case '1':
      firmware1_file = optarg;
      break;
    case '2':
      firmware2_file = optarg;
      break;
    case 'r':
      rockbox_file = optarg;
      break;
    default:
      abort();
    }
  }

  if (log_file) {
    log_fp = fopen(log_file, "w");
    if (log_fp)
      log_set_fp(log_fp);
  }
}

int main(int argc, char *argv[]) {
  int upgrade = 1;

  program_name = argv[0];
  if (parse_options(argc, argv) == -1)
    return 1;

  init_events();

  memory = create_memory();
  if (!memory)
    return 1;

  cpu = create_cpu(memory);
  if (!cpu)
    return 1;

  hardware = create_hardware();
  if (!hardware)
    return 1;

  cpu_set_irq_p(cpu, check_irq, hardware);

  hardware_reset(hardware);
  install_hardware(hardware, memory);

  if (init_iriver(cpu, memory, 0) < 0)
    return 1;

  trace_detect_firmware(memory);
  trace_init();
  if (rockbox_file)
    stack_tracing = 0;

  trainer_init(ram4->mem);

  if (option_disable_ui || !ui_init()) {
    if (option_gdb)
      gdb_target();
    else
      go(0, NULL, 0);

    if (!option_disable_ui)
      ui_cleanup();
  }

  return 0;
}
