/* iRiver iFP 7XX Emulator
   Copyright (C) 2005 Tomasz Malesinski <tmal@mimuw.edu.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <stdlib.h>

#include "emuriver.h"
#include "memory.h"
#include "cpu.h"
#include "log.h"

uint8 read_void(void *data, uint32 adr) {
  logmsg(LOG_MEMORY, 5, "void read from %8x\n", adr);
  return 0xff;
}

uint32 read_half_word_void(void *data, uint32 adr) {
  logmsg(LOG_MEMORY, 5, "void read from %8x (pc=%8x)\n", adr, cpu->reg[15] - 4);
  return 0xffff;
}

uint32 read_word_void(void *data, uint32 adr) {
  logmsg(LOG_MEMORY, 5, "void read from %8x (pc=%8x)\n", adr, cpu->reg[15] - 4);
  return 0xffffffff;
}

void write_void(void *data, uint32 adr, uint8 val) {
  logmsg(LOG_MEMORY, 5, "void write %8x to %8x\n", val, adr);
}

void write_half_word_void(void *data, uint32 adr, uint32 val) {
  logmsg(LOG_MEMORY, 5, "void write %8x to %8x\n", val, adr);
}

void write_word_void(void *data, uint32 adr, uint32 val) {
  logmsg(LOG_MEMORY, 5, "void write %8x to %8x\n", val, adr);
}

static void rom_write(void *data, uint32 adr, uint8 val) {
  logmsg(LOG_MEMORY, 5, "write %8x to ROM at %8x\n", val, adr);
}

static void rom_write_half_word(void *data, uint32 adr, uint32 val) {
  logmsg(LOG_MEMORY, 5, "write %8x to ROM at %8x\n", val, adr);
}

static void rom_write_word(void *data, uint32 adr, uint32 val) {
  logmsg(LOG_MEMORY, 5, "write %8x to ROM at %8x\n", val, adr);
}

static uint8 block_read(void *data, uint32 adr) {
  struct mem_block *b = data;
  return b->mem[adr - b->base];
}

static uint32 block_read_half_word(void *data, uint32 adr) {
  struct mem_block *b = data;
  uint32 r = 0;
  int i;
  for (i = 1; i >= 0; i--)
    r = (r << 8) | b->mem[adr - b->base + i];
  return r;
}

static uint32 block_read_word(void *data, uint32 adr) {
  struct mem_block *b = data;
  uint32 r = 0;
  int i;
  for (i = 3; i >= 0; i--)
    r = (r << 8) | b->mem[adr - b->base + i];
  return r;
}

static void block_write(void *data, uint32 adr, uint8 val) {
  struct mem_block *b = data;
  b->mem[adr - b->base] = val;
}

static void block_write_half_word(void *data, uint32 adr, uint32 val) {
  struct mem_block *b = data;
  int i;
  for (i = 0; i < 2; i++)
    b->mem[adr - b->base + i] = (val >> (i * 8)) & 0xff;
}

static void block_write_word(void *data, uint32 adr, uint32 val) {
  struct mem_block *b = data;
  int i;
  for (i = 0; i < 4; i++)
    b->mem[adr - b->base + i] = (val >> (i * 8)) & 0xff;
}

uint8 memory_read(struct memory *memory, uint32 adr) {
  struct mem_region *r = &(memory->region[adr >> 16]);
  return (*(r->read))(r->data, adr);
}

void memory_write(struct memory *memory, uint32 adr, uint8 val) {
  struct mem_region *r = &(memory->region[adr >> 16]);
  (*(r->write))(r->data, adr, val);
}

uint32 memory_read_half_word(struct memory *memory, uint32 adr) {
  struct mem_region *r = &(memory->region[adr >> 16]);
  return (*(r->read_half_word))(r->data, adr);
}

void memory_write_half_word(struct memory *memory, uint32 adr, uint32 val) {
  struct mem_region *r = &(memory->region[adr >> 16]);
  (*(r->write_half_word))(r->data, adr, val);
}

uint32 memory_read_word(struct memory *memory, uint32 adr) {
  struct mem_region *r = &(memory->region[adr >> 16]);
  return (*(r->read_word))(r->data, adr);
}

void memory_write_word(struct memory *memory, uint32 adr, uint32 val) {
  struct mem_region *r = &(memory->region[adr >> 16]);
  (*(r->write_word))(r->data, adr, val);
}

void set_area(struct memory *mem, uint32 start, uint32 end,
	      uint8 (*read)(void *, uint32),
	      void (*write)(void *, uint32, uint8),
	      uint32 (*read_half_word)(void *, uint32),
	      void (*write_half_word)(void *, uint32, uint32),
	      uint32 (*read_word)(void *, uint32),
	      void (*write_word)(void *, uint32, uint32),
	      void *data) {
  int r;
  for (r = start >> 16; r <= end >> 16; r++) {
    mem->region[r].read = read;
    mem->region[r].write = write;
    mem->region[r].read_half_word = read_half_word;
    mem->region[r].write_half_word = write_half_word;
    mem->region[r].read_word = read_word;
    mem->region[r].write_word = write_word;
    mem->region[r].data = data;
  }    
}

struct memory *create_memory(void) {
  struct memory *m;
  int i;

  m = xmalloc(sizeof(struct memory));
  if (!m)
    return NULL;
  
  for (i = 0; i < (1 << 16); i++) {
    m->region[i].read = read_void;
    m->region[i].write = write_void;
    m->region[i].read_half_word = read_half_word_void;
    m->region[i].write_half_word = write_half_word_void;
    m->region[i].read_word = read_word_void;
    m->region[i].write_word = write_word_void;
  }

  return m;
}

static struct mem_block *create_memory_block(uint32 adr, uint32 len) {
  uint8 *area;
  struct mem_block *block;

  area = xmalloc(len);
  if (!area) return NULL;

  block = xmalloc(sizeof(struct mem_block));
  if (!block) {
    free(area);
    return NULL;
  }

  block->mem = area;
  block->base = adr;
  block->len = len;

  return block;
}

void free_memory_block(struct mem_block *block) {
  free(block->mem);
  free(block);
}

struct mem_block *create_memory_area_rom(struct memory *mem, uint32 adr,
					 uint32 len, char *file) {
  struct mem_block *block;
  FILE *fp;

  block = create_memory_block(adr, len);
  if (!block) return NULL;

  fp = fopen(file, "rb");
  if (!fp) {
    error_sys("%s: ", file);
    free_memory_block(block);
    return NULL;
  }

  fread(block->mem, 1, len, fp);
  fclose(fp);

  block->rom = 1;

  set_area(mem, adr, adr + len - 1,
	   block_read, rom_write,
	   block_read_half_word, rom_write_half_word,
	   block_read_word, rom_write_word,
	   block);

  return block;
}

struct mem_block *create_memory_area_ram(struct memory *mem, uint32 adr,
					 uint32 len) {
  struct mem_block *block;
  FILE *fp;

  block = create_memory_block(adr, len);
  if (!block)
    return NULL;

  block->rom = 0;
  memset(block->mem, 0, len);

  set_area(mem, adr, adr + len - 1,
	   block_read, block_write,
	   block_read_half_word, block_write_half_word,
	   block_read_word, block_write_word,
	   block);

  return block;
}
